@@ -3,7 +3,7 @@
 """
 DOM wrapper around libxml2, specifically the libxml2mod Python extension module.
 
-Copyright (C) 2003, 2004, 2005, 2006, 2007 Paul Boddie <paul@boddie.org.uk>
+Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Paul Boddie <paul@boddie.org.uk>
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
@@ -19,7 +19,7 @@ You should have received a copy of the GNU Lesser General Public License along
 with this program.  If not, see <http://www.gnu.org/licenses/>.
 """
 
-__version__ = "0.4.5"
+__version__ = "0.4.7"
 
 from libxml2dom.macrolib import *
 from libxml2dom.macrolib import \
@@ -29,7 +29,7 @@ from libxml2dom.macrolib import \
     toString as Node_toString, toStream as Node_toStream, \
     toFile as Node_toFile
 import urllib # for parseURI in HTML mode
-import xml.dom # for getElementById
+import libxml2dom.errors
 
 # Standard namespaces.
 
@@ -59,10 +59,21 @@ class Implementation(object):
     # Factory functions.
 
     def get_node(self, _node, context_node):
+
+        # Return the existing document.
+
         if Node_nodeType(_node) == context_node.DOCUMENT_NODE:
             return context_node.ownerDocument
+
+        # Return an attribute using the parent of the attribute as the owner
+        # element.
+
         elif Node_nodeType(_node) == context_node.ATTRIBUTE_NODE:
-            return Attribute(_node, self, context_node.ownerDocument, context_node)
+            return Attribute(_node, self, context_node.ownerDocument,
+                self.get_node(Node_parentNode(_node), context_node))
+
+        # Return other nodes.
+
         else:
             return Node(_node, self, context_node.ownerDocument)
 
@@ -382,9 +393,9 @@ class Node(object):
 
     def insertBefore(self, tmp, oldNode):
         if tmp.ownerDocument != self.ownerDocument:
-            raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
+            raise xml.dom.WrongDocumentErr()
         if oldNode.parentNode != self:
-            raise xml.dom.DOMException(xml.dom.NOT_FOUND_ERR)
+            raise xml.dom.NotFoundErr()
         if hasattr(tmp, "as_native_node"):
             return self.impl.get_node(Node_insertBefore(self._node, tmp.as_native_node(), oldNode.as_native_node()), self)
         else:
@@ -392,9 +403,9 @@ class Node(object):
 
     def replaceChild(self, tmp, oldNode):
         if tmp.ownerDocument != self.ownerDocument:
-            raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
+            raise xml.dom.WrongDocumentErr()
         if oldNode.parentNode != self:
-            raise xml.dom.DOMException(xml.dom.NOT_FOUND_ERR)
+            raise xml.dom.NotFoundErr()
         if hasattr(tmp, "as_native_node"):
             return self.impl.get_node(Node_replaceChild(self._node, tmp.as_native_node(), oldNode.as_native_node()), self)
         else:
@@ -402,7 +413,7 @@ class Node(object):
 
     def appendChild(self, tmp):
         if tmp.ownerDocument != self.ownerDocument:
-            raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
+            raise xml.dom.WrongDocumentErr()
         if hasattr(tmp, "as_native_node"):
             return self.impl.get_node(Node_appendChild(self._node, tmp.as_native_node()), self)
         else:
@@ -416,12 +427,11 @@ class Node(object):
         return tmp
 
     def getElementById(self, identifier):
-        nodes = self.xpath(".//*[@xml:id='" + identifier.replace("'", "&apos;") + "']",
-            namespaces={"xml" : xml.dom.XML_NAMESPACE})
-        if nodes:
-            return nodes[0]
-        else:
+        _node = Node_getElementById(self.ownerDocument.as_native_node(), identifier)
+        if _node is None:
             return None
+        else:
+            return self.impl.get_node(_node, self)
 
     def getElementsByTagName(self, tagName):
         return self.xpath(".//" + tagName)
@@ -488,6 +498,12 @@ class Node(object):
     # NOTE: To be finished.
 
     def xpath(self, expr, variables=None, namespaces=None):
+
+        """
+        Evaluate the given expression 'expr' using the optional 'variables' and
+        'namespaces' mappings.
+        """
+
         ns = {}
         ns.update(default_ns)
         ns.update(namespaces or {})
@@ -499,6 +515,18 @@ class Node(object):
         else:
             return result
 
+    # Other extensions to the usual PyXML API.
+
+    def xinclude(self):
+
+        """
+        Process XInclude declarations within the document, returning the number
+        of substitutions performed (zero or more), raising an XIncludeException
+        otherwise.
+        """
+
+        return Node_xinclude(self._node)
+
     # Convenience methods.
 
     def toString(self, encoding=None, prettyprint=0):
@@ -531,12 +559,23 @@ class _Document:
 
     """
     An abstract class providing document-level housekeeping and distinct
-    functionality.
+    functionality. Configuration of the document is also supported.
+    See: http://www.w3.org/TR/DOM-Level-3-Core/core.html#DOMConfiguration
     """
 
+    # Constants from 
+    # See: http://www.w3.org/TR/DOM-Level-3-Val/validation.html#VAL-Interfaces-NodeEditVAL
+
+    VAL_TRUE = 5
+    VAL_FALSE = 6
+    VAL_UNKNOWN = 7
+
     def __init__(self, node, impl):
         self._node = node
         self.implementation = self.impl = impl
+        self.error_handler = libxml2dom.errors.DOMErrorHandler()
+
+    # Standard DOM properties and their implementations.
 
     def _documentElement(self):
         return self.xpath("*")[0]
@@ -551,6 +590,55 @@ class _Document:
     documentElement = property(_documentElement)
     ownerDocument = property(_ownerDocument)
 
+    # DOM Level 3 Core DOMConfiguration methods.
+
+    def setParameter(self, name, value):
+        if name == "error-handler":
+            raise xml.dom.NotSupportedErr()
+        raise xml.dom.NotFoundErr()
+
+    def getParameter(self, name):
+        if name == "error-handler":
+            return self.error_handler
+        raise xml.dom.NotFoundErr()
+
+    def canSetParameter(self, name, value):
+        return 0
+
+    def _parameterNames(self):
+        return []
+
+    # Extensions to the usual PyXML API.
+
+    def validate(self, doc):
+
+        """
+        Validate the document against the given schema document, 'doc'.
+        """
+
+        validation_ns = doc.documentElement.namespaceURI
+
+        if hasattr(doc, "as_native_node"):
+            _schema = Document_schema(doc.as_native_node(), validation_ns)
+        else:
+            _schema = Document_schemaFromString(doc.toString(), validation_ns)
+        try:
+            self.error_handler.reset()
+            return Document_validate(_schema, self._node, self.error_handler, validation_ns)
+        finally:
+            Schema_free(_schema, validation_ns)
+
+    # DOM Level 3 Validation methods.
+
+    def validateDocument(self, doc):
+
+        """
+        Validate the document against the given schema document, 'doc'.
+        See: http://www.w3.org/TR/DOM-Level-3-Val/validation.html#VAL-Interfaces-DocumentEditVAL-validateDocument
+        """
+
+        return self.validate(doc) and self.VAL_TRUE or self.VAL_FALSE
+
 class Document(_Document, Node):
 
     """
@@ -590,7 +678,7 @@ def createDocumentType(localName, publicId, systemId):
 def createDocument(namespaceURI, localName, doctype):
     return default_impl.createDocument(namespaceURI, localName, doctype)
 
-def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0, impl=None):
+def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0, validate=0, remote=0, impl=None):
 
     """
     Parse the given 'stream_or_string', where the supplied object can either be
@@ -607,6 +695,13 @@ def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0, impl=None):
     documents will be parsed, even though such documents may be missing content
     such as closing tags.
 
+    If the optional 'validate' parameter is set to a true value, an attempt will
+    be made to validate the parsed document.
+
+    If the optional 'remote' parameter is set to a true value, references to
+    remote documents (such as DTDs) will be followed in order to obtain such
+    documents.
+
     A document object is returned by this function.
     """
 
@@ -614,11 +709,13 @@ def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0, impl=None):
 
     if hasattr(stream_or_string, "read"):
         stream = stream_or_string
-        return parseString(stream.read(), html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=impl)
+        return parseString(stream.read(), html=html, htmlencoding=htmlencoding,
+            unfinished=unfinished, validate=validate, remote=remote, impl=impl)
     else:
-        return parseFile(stream_or_string, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=impl)
+        return parseFile(stream_or_string, html=html, htmlencoding=htmlencoding,
+            unfinished=unfinished, validate=validate, remote=remote, impl=impl)
 
-def parseFile(filename, html=0, htmlencoding=None, unfinished=0, impl=None):
+def parseFile(filename, html=0, htmlencoding=None, unfinished=0, validate=0, remote=0, impl=None):
 
     """
     Parse the file having the given 'filename'. The optional parameters
@@ -633,13 +730,21 @@ def parseFile(filename, html=0, htmlencoding=None, unfinished=0, impl=None):
     documents will be parsed, even though such documents may be missing content
     such as closing tags.
 
+    If the optional 'validate' parameter is set to a true value, an attempt will
+    be made to validate the parsed document.
+
+    If the optional 'remote' parameter is set to a true value, references to
+    remote documents (such as DTDs) will be followed in order to obtain such
+    documents.
+
     A document object is returned by this function.
     """
 
     impl = impl or default_impl
-    return impl.adoptDocument(Node_parseFile(filename, html=html, htmlencoding=htmlencoding, unfinished=unfinished))
+    return impl.adoptDocument(Node_parseFile(filename, html=html, htmlencoding=htmlencoding,
+        unfinished=unfinished, validate=validate, remote=remote))
 
-def parseString(s, html=0, htmlencoding=None, unfinished=0, impl=None):
+def parseString(s, html=0, htmlencoding=None, unfinished=0, validate=0, remote=0, impl=None):
 
     """
     Parse the content of the given string 's'. The optional parameters described
@@ -654,13 +759,21 @@ def parseString(s, html=0, htmlencoding=None, unfinished=0, impl=None):
     documents will be parsed, even though such documents may be missing content
     such as closing tags.
 
+    If the optional 'validate' parameter is set to a true value, an attempt will
+    be made to validate the parsed document.
+
+    If the optional 'remote' parameter is set to a true value, references to
+    remote documents (such as DTDs) will be followed in order to obtain such
+    documents.
+
     A document object is returned by this function.
     """
 
     impl = impl or default_impl
-    return impl.adoptDocument(Node_parseString(s, html=html, htmlencoding=htmlencoding, unfinished=unfinished))
+    return impl.adoptDocument(Node_parseString(s, html=html, htmlencoding=htmlencoding,
+        unfinished=unfinished, validate=validate, remote=remote))
 
-def parseURI(uri, html=0, htmlencoding=None, unfinished=0, impl=None):
+def parseURI(uri, html=0, htmlencoding=None, unfinished=0, validate=0, remote=0, impl=None):
 
     """
     Parse the content found at the given 'uri'. The optional parameters
@@ -675,6 +788,13 @@ def parseURI(uri, html=0, htmlencoding=None, unfinished=0, impl=None):
     documents will be parsed, even though such documents may be missing content
     such as closing tags.
 
+    If the optional 'validate' parameter is set to a true value, an attempt will
+    be made to validate the parsed document.
+
+    If the optional 'remote' parameter is set to a true value, references to
+    remote documents (such as DTDs) will be followed in order to obtain such
+    documents.
+
     XML documents are retrieved using libxml2's own network capabilities; HTML
     documents are retrieved using the urllib module provided by Python. To
     retrieve either kind of document using Python's own modules for this purpose
@@ -692,12 +812,14 @@ def parseURI(uri, html=0, htmlencoding=None, unfinished=0, impl=None):
     if html:
         f = urllib.urlopen(uri)
         try:
-            return parse(f, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=impl)
+            return parse(f, html=html, htmlencoding=htmlencoding, unfinished=unfinished,
+                validate=validate, remote=remote, impl=impl)
         finally:
             f.close()
     else:
         impl = impl or default_impl
-        return impl.adoptDocument(Node_parseURI(uri, html=html, htmlencoding=htmlencoding, unfinished=unfinished))
+        return impl.adoptDocument(Node_parseURI(uri, html=html, htmlencoding=htmlencoding,
+            unfinished=unfinished, validate=validate, remote=remote))
 
 def toString(node, encoding=None, prettyprint=0):
 
