@@ -3,7 +3,7 @@
 """
 DOM macros for virtual libxml2mod node methods and properties.
 
-Copyright (C) 2003, 2004, 2005, 2006, 2007 Paul Boddie <paul@boddie.org.uk>
+Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Paul Boddie <paul@boddie.org.uk>
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
@@ -20,6 +20,7 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 """
 
 import xml.dom
+from libxml2dom.errors import DOMError
 
 # Try the conventional import first.
 
@@ -458,7 +459,7 @@ def Node_importNode(node, other, deep):
     elif Node_nodeType(other) == xml.dom.Node.CDATA_SECTION_NODE:
         return Node_createCDATASection(node, Node_data(other))
 
-    raise xml.dom.DOMException(xml.dom.NOT_SUPPORTED_ERR, "Node type '%s' (%d) not supported." % (other, Node_nodeType(other)))
+    raise xml.dom.NotSupportedErr("Node type '%s' (%d) not supported." % (other, Node_nodeType(other)))
 
 def Node_importNode_DOM(node, other, deep):
     if other.nodeType == xml.dom.Node.ELEMENT_NODE:
@@ -483,7 +484,16 @@ def Node_importNode_DOM(node, other, deep):
     elif other.nodeType == xml.dom.Node.CDATA_SECTION_NODE:
         return Node_createCDATASection(node, other.data)
 
-    raise xml.dom.DOMException(xml.dom.NOT_SUPPORTED_ERR, "Node type '%s' (%d) not supported." % (_reverseNodeTypes[other.nodeType], other.nodeType))
+    raise xml.dom.NotSupportedErr(
+        "Node type '%s' (%d) not supported." % (_reverseNodeTypes[other.nodeType], other.nodeType)
+        )
+
+def Node_getElementById(doc, identifier):
+    node = libxml2mod.xmlGetID(doc, identifier)
+    if node is None:
+        return None
+    else:
+        return Node_parentNode(node)
 
 def Node_xpath(node, expr, variables=None, namespaces=None):
     expr = from_unicode(expr)
@@ -503,6 +513,13 @@ def Node_xpath(node, expr, variables=None, namespaces=None):
     libxml2mod.xmlXPathFreeContext(context)
     return result
 
+def Node_xinclude(node):
+    result = libxml2mod.xmlXIncludeProcessFlags(node, XML_PARSE_NOERROR | XML_PARSE_NOWARNING | XML_PARSE_NONET)
+    if result == -1:
+        raise XIncludeException()
+    else:
+        return result
+
 # Exceptions.
 
 class LSException(Exception):
@@ -513,16 +530,23 @@ class LSException(Exception):
     SERIALIZE_ERR = 82
 
     def __repr__(self):
-        return str(self)
+        exctype, excdata = self.args[0:2]
+        return "LSException(%d, %r)" % (exctype, excdata)
 
     def __str__(self):
-        exctype = self.args[0]
+        exctype, excdata = self.args[0:2]
         if exctype == self.PARSE_ERR:
-            return "Parse error: LSException(%d)" % exctype
+            return "Parse error: %s" % excdata
         elif exctype == self.SERIALIZE_ERR:
-            return "Serialize error: LSException(%d)" % exctype
+            return "Serialize error: %s" % excdata
         else:
-            return Exception.__repr__(self)
+            return repr(self)
+
+class XIncludeException(Exception):
+
+    "Unstandardised XInclude exception."
+
+    pass
 
 # Utility functions.
 
@@ -544,57 +568,76 @@ def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0):
     else:
         return parseFile(stream_or_string, html=html, htmlencoding=htmlencoding, unfinished=unfinished)
 
-def parseFile(s, html=0, htmlencoding=None, unfinished=0):
-    # NOTE: Switching off validation and remote DTD resolution.
+def parseFile(s, html=0, htmlencoding=None, unfinished=0, validate=0, remote=0):
     if not html:
         context = libxml2mod.xmlCreateFileParserCtxt(s)
-        if context is None:
-            raise LSException(LSException.PARSE_ERR)
-        Parser_configure(context)
-        Parser_parse(context)
-        doc = Parser_document(context)
-        if unfinished or Parser_well_formed(context):
-            return doc
-        else:
-            raise LSException(LSException.PARSE_ERR)
+        return _parseXML(context, unfinished, validate, remote)
     else:
-        return libxml2mod.htmlReadFile(s, htmlencoding, HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET)
+        return libxml2mod.htmlReadFile(s, htmlencoding,
+            HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | html_net_flag(remote))
 
-def parseString(s, html=0, htmlencoding=None, unfinished=0):
-    # NOTE: Switching off validation and remote DTD resolution.
+def parseString(s, html=0, htmlencoding=None, unfinished=0, validate=0, remote=0):
     if not html:
         context = libxml2mod.xmlCreateMemoryParserCtxt(s, len(s))
-        if context is None:
-            raise LSException(LSException.PARSE_ERR)
-        Parser_configure(context)
-        Parser_parse(context)
-        doc = Parser_document(context)
-        if unfinished or Parser_well_formed(context):
-            return doc
-        else:
-            raise LSException(LSException.PARSE_ERR)
+        return _parseXML(context, unfinished, validate, remote)
     else:
         # NOTE: URL given as None.
         html_url = None
         return libxml2mod.htmlReadMemory(s, len(s), html_url, htmlencoding,
-            HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET)
+            HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | html_net_flag(remote))
 
-def parseURI(uri, html=0, htmlencoding=None, unfinished=0):
-    # NOTE: Switching off validation and remote DTD resolution.
+def parseURI(uri, html=0, htmlencoding=None, unfinished=0, validate=0, remote=0):
     if not html:
         context = libxml2mod.xmlCreateURLParserCtxt(uri, 0)
-        if context is None:
-            raise LSException(LSException.PARSE_ERR)
-        Parser_configure(context)
-        Parser_parse(context)
-        doc = Parser_document(context)
-        if unfinished or Parser_well_formed(context):
-            return doc
-        else:
-            raise LSException(LSException.PARSE_ERR)
+        return _parseXML(context, unfinished, validate, remote)
     else:
         raise NotImplementedError, "parseURI does not yet support HTML"
 
+def _parseXML(context, unfinished, validate, remote):
+    if context is None:
+        raise LSException(LSException.PARSE_ERR, DOMError(DOMError.SEVERITY_FATAL_ERROR))
+
+    Parser_configure(context, validate, remote)
+    Parser_parse(context)
+    doc = Parser_document(context)
+    error = Parser_error()
+
+    try:
+        if validate and not Parser_valid(context):
+
+            # NOTE: May not be the correct exception.
+
+            raise LSException(
+                LSException.PARSE_ERR,
+                DOMError(
+                    DOMError.SEVERITY_FATAL_ERROR,
+                    get_parse_error_message() or "Document did not validate"
+                    ))
+
+        elif unfinished and (error is None or Parser_errorCode(error) == XML_ERR_TAG_NOT_FINISHED):
+
+            # NOTE: There may be other unfinished conditions.
+
+            return doc
+
+        elif error is not None and Parser_errorLevel(error) == XML_ERR_FATAL:
+            raise LSException(
+                LSException.PARSE_ERR,
+                DOMError(
+                    DOMError.SEVERITY_FATAL_ERROR,
+                    get_parse_error_message() or "Document caused fatal error"
+                    ))
+
+        else:
+
+            # NOTE: Could provide non-fatal errors or warnings.
+
+            return doc
+
+    finally:
+        Parser_resetError(error)
+        libxml2mod.xmlFreeParserCtxt(context)
+
 def toString(node, encoding=None, prettyprint=0):
     return libxml2mod.serializeNode(node, encoding, prettyprint)
 
@@ -609,17 +652,73 @@ def toFile(node, f, encoding=None, prettyprint=0):
 HTML_PARSE_NOERROR = 32
 HTML_PARSE_NOWARNING = 64
 HTML_PARSE_NONET = 2048
+XML_PARSE_DTDVALID = 16
 XML_PARSE_NOERROR = 32
 XML_PARSE_NOWARNING = 64
 XML_PARSE_NONET = 2048
 
+XML_ERR_NONE = 0
+XML_ERR_WARNING = 1
+XML_ERR_ERROR = 2
+XML_ERR_FATAL = 3
+
+XML_ERR_TAG_NOT_FINISHED = 77
+
+def html_net_flag(remote):
+    if remote:
+        return 0
+    else:
+        return HTML_PARSE_NONET
+
+def xml_net_flag(remote):
+    if remote:
+        return 0
+    else:
+        return XML_PARSE_NONET
+
+def xml_validate_flag(validate):
+    if validate:
+        return XML_PARSE_DTDVALID
+    else:
+        return 0
+
+def get_parse_error_message():
+    error = Parser_error()
+    if error is not None:
+        filename = libxml2mod.xmlErrorGetFile(error)
+        if filename is None:
+            filename = "<string>"
+        else:
+            filename = repr(filename)
+        line = libxml2mod.xmlErrorGetLine(error)
+        error_message = libxml2mod.xmlErrorGetMessage(error).strip()
+        return "Filename %s, line %d: %s" % (filename, line, error_message)
+    else:
+        return None
+
+def Parser_error():
+    return libxml2mod.xmlGetLastError()
+
+def Parser_resetError(error):
+    if error is None:
+        return libxml2mod.xmlResetLastError()
+    else:
+        return libxml2mod.xmlResetError(error)
+
+def Parser_errorLevel(error):
+    return libxml2mod.xmlErrorGetLevel(error)
+
+def Parser_errorCode(error):
+    return libxml2mod.xmlErrorGetCode(error)
+
 def Parser_push():
     return libxml2mod.xmlCreatePushParser(None, "", 0, None)
 
-def Parser_configure(context):
+def Parser_configure(context, validate, remote):
     libxml2mod.xmlParserSetPedantic(context, 0)
-    libxml2mod.xmlParserSetValidate(context, 0)
-    libxml2mod.xmlCtxtUseOptions(context, XML_PARSE_NOERROR | XML_PARSE_NOWARNING | XML_PARSE_NONET)
+    #libxml2mod.xmlParserSetValidate(context, validate)
+    libxml2mod.xmlCtxtUseOptions(context,
+        XML_PARSE_NOERROR | XML_PARSE_NOWARNING | xml_net_flag(remote) | xml_validate_flag(validate))
 
 def Parser_feed(context, s):
     libxml2mod.xmlParseChunk(context, s, len(s), 1)
@@ -627,10 +726,122 @@ def Parser_feed(context, s):
 def Parser_well_formed(context):
     return libxml2mod.xmlParserGetWellFormed(context)
 
+def Parser_valid(context):
+    return libxml2mod.xmlParserGetIsValid(context)
+
 def Parser_document(context):
     return libxml2mod.xmlParserGetDoc(context)
 
 def Parser_parse(context):
     libxml2mod.xmlParseDocument(context)
 
+# Schema and validation helper functions and classes.
+# NOTE: Should potentially combine these with other definitions.
+
+RELAXNG_NS = "http://relaxng.org/ns/structure/1.0"
+SCHEMATRON_NS = "http://purl.oclc.org/dsdl/schematron"
+XMLSCHEMA_NS = "http://www.w3.org/2001/XMLSchema"
+
+def Document_schema(doc, namespaceURI):
+    if namespaceURI == RELAXNG_NS:
+        return Schema_parseRelaxNG(libxml2mod.xmlRelaxNGNewDocParserCtxt(doc))
+    elif namespaceURI == SCHEMATRON_NS:
+        return Schema_parseSchematron(libxml2mod.xmlSchematronNewDocParserCtxt(doc))
+    elif namespaceURI == XMLSCHEMA_NS:
+        return Schema_parseSchema(libxml2mod.xmlSchemaNewDocParserCtxt(doc))
+    else:
+        return None
+
+def Document_schemaFromString(s, namespaceURI):
+    if namespaceURI == RELAXNG_NS:
+        return Schema_parseRelaxNG(libxml2mod.xmlRelaxNGNewMemParserCtxt(s, len(s)))
+    elif namespaceURI == SCHEMATRON_NS:
+        return Schema_parseSchematron(libxml2mod.xmlSchematronNewMemParserCtxt(s, len(s)))
+    elif namespaceURI == XMLSCHEMA_NS:
+        return Schema_parseSchema(libxml2mod.xmlSchemaNewMemParserCtxt(s, len(s)))
+    else:
+        return None
+
+def Document_validate(schema, doc, error_handler, namespaceURI):
+    if namespaceURI == RELAXNG_NS:
+        return Document_validateRelaxNG(schema, doc, error_handler)
+    elif namespaceURI == SCHEMATRON_NS:
+        return Document_validateSchematron(schema, doc, error_handler)
+    elif namespaceURI == XMLSCHEMA_NS:
+        return Document_validateSchema(schema, doc, error_handler)
+    else:
+        return 0
+
+def Document_validateRelaxNG(schema, doc, error_handler):
+    validator_context = libxml2mod.xmlRelaxNGNewValidCtxt(schema)
+    handler = ValidationHandler(error_handler)
+    libxml2mod.xmlRelaxNGSetValidErrors(validator_context, handler.error, handler.warning, None)
+    try:
+        status = libxml2mod.xmlRelaxNGValidateDoc(validator_context, doc)
+        return status == 0
+    finally:
+        libxml2mod.xmlRelaxNGFreeValidCtxt(validator_context)
+
+def Document_validateSchematron(schema, doc, error_handler):
+    validator_context = libxml2mod.xmlSchematronNewValidCtxt(schema)
+    handler = ValidationHandler(error_handler)
+    libxml2mod.xmlSchematronSetValidErrors(validator_context, handler.error, handler.warning, None)
+    try:
+        status = libxml2mod.xmlSchematronValidateDoc(validator_context, doc)
+        return status == 0
+    finally:
+        libxml2mod.xmlSchematronFreeValidCtxt(validator_context)
+
+def Document_validateSchema(schema, doc, error_handler):
+    validator_context = libxml2mod.xmlSchemaNewValidCtxt(schema)
+    handler = ValidationHandler(error_handler)
+    libxml2mod.xmlSchemaSetValidErrors(validator_context, handler.error, handler.warning, None)
+    try:
+        status = libxml2mod.xmlSchemaValidateDoc(validator_context, doc)
+        return status == 0
+    finally:
+        libxml2mod.xmlSchemaFreeValidCtxt(validator_context)
+
+def Schema_parseRelaxNG(context):
+    try:
+        return libxml2mod.xmlRelaxNGParse(context)
+    finally:
+        libxml2mod.xmlRelaxNGFreeParserCtxt(context)
+
+def Schema_parseSchematron(context):
+    try:
+        return libxml2mod.xmlSchematronParse(context)
+    finally:
+        libxml2mod.xmlSchematronFreeParserCtxt(context)
+
+def Schema_parseSchema(context):
+    try:
+        return libxml2mod.xmlSchemaParse(context)
+    finally:
+        libxml2mod.xmlSchemaFreeParserCtxt(context)
+
+def Schema_free(schema, namespaceURI):
+    if namespaceURI == RELAXNG_NS:
+        libxml2mod.xmlRelaxNGFree(schema)
+    elif namespaceURI == SCHEMATRON_NS:
+        libxml2mod.xmlSchematronFree(schema)
+    elif namespaceURI == XMLSCHEMA_NS:
+        libxml2mod.xmlSchemaFree(schema)
+
+class ValidationHandler:
+
+    """
+    A handler which collects validation errors and warnings and passes them to a
+    DOMErrorHandler.
+    """
+
+    def __init__(self, error_handler):
+        self.error_handler = error_handler
+
+    def error(self, msg, arg):
+        self.error_handler.handleError(DOMError(DOMError.SEVERITY_FATAL_ERROR, msg.strip()))
+
+    def warning(self, msg, arg):
+        self.error_handler.handleError(DOMError(DOMError.SEVERITY_WARNING, msg.strip()))
+
 # vim: tabstop=4 expandtab shiftwidth=4
